#!/usr/bin/perl -w
#
# file    : index.pl
# changed : 2005-12-28
# author  : Bas Couwenberg <sebastic@xs4all.nl>
#

use strict qw(subs vars);
use Cwd;
use File::Basename ();
use Date::Calc qw(:all);
use Apache::Request;
use Image::Magick;

my $r = Apache->request;
my $c = $r->connection;
my $s = $r->server;

my $cwd = Cwd::cwd.'/';

# Initialize global variables
my %args = &get_args($r, $c, $s);
   $args{'fci-version'} 	 	= '0.10';
   $args{'www-data-dir'}		= $cwd;
   $args{'stanford-data-dir'}	 	= 'stanford-data/';
   $args{'client-data-dir'}	 	= 'client-data/';
   $args{'site-data-dir'}		= 'settings/site-data/';
   $args{'img-dir'}			= 'img/';
   $args{'bin-dir'}			= 'bin/';
  
   $args{'client-list-file'}	 	= $args{'client-data-dir'}.'client-list';
   $args{'project-list-file'}	 	= $args{'client-data-dir'}.'project-list';
   $args{'username-list-file'}	 	= $args{'client-data-dir'}.'username-list';

   $args{'user-summary-file'}	 	= $args{'stanford-data-dir'}.'daily_user_summary.txt';
   $args{'team-summary-file'}	 	= $args{'stanford-data-dir'}.'daily_team_summary.txt';
   $args{'project-summary-file'} 	= $args{'stanford-data-dir'}.'psummaryC.html';
   
   $args{'user-summary-update-file'}	= $args{'user-summary-file'}.'.lastupdate';
   $args{'team-summary-update-file'}	= $args{'team-summary-file'}.'.lastupdate';
   $args{'project-summary-update-file'} = $args{'project-summary-file'}.'.lastupdate';
   
   $args{'qd-files-dir'}		= 'qd-files/';
   
   #$args{'qd-base-url'}                = 'http://www.boston.quik.com/rph/';
   #$args{'qd-url'}			= 'http://www.boston.quik.com/rph/fah.html#qd';
   #$args{'qdinfo-url'}			= 'http://www.boston.quik.com/rph/fah.html#qdinfo';
   $args{'qd-base-url'}                 = 'http://linuxminded.xs4all.nl/mirror/www.boston.quik.com/rph/';
   $args{'qd-url'}			= 'http://linuxminded.xs4all.nl/mirror/www.boston.quik.com/rph/fah.html#qd';
   $args{'qdinfo-url'}			= 'http://linuxminded.xs4all.nl/mirror/www.boston.quik.com/rph/fah.html#qdinfo';
   $args{'xyz2pdb-url'}                 = 'http://linuxminded.xs4all.nl/mirror/www.boston.quik.com/rph/fah.html#xyz2pdb';
   
   $args{'qd-bin-linux-url'}            = 'http://linuxminded.xs4all.nl/software/qd-tools/binaries/linux/qd';
   $args{'qd-bin-win32-url'}            = 'http://linuxminded.xs4all.nl/software/qd-tools/binaries/win32/qd.exe';
   $args{'qdinfo.dat-url'}              = 'http://linuxminded.xs4all.nl/software/qd-tools/source/qdinfo/qdinfo.dat';

   $args{'rasmol-bin'}			= $args{'bin-dir'}.'rasmol_32BIT';
   $args{'xyz2pdb-bin'}			= $args{'bin-dir'}.'xyz2pdb';

   $args{'convert-bin'}                 = '/usr/bin/convert';

   $args{'site-config-file'}		= $args{'site-data-dir'}.'config';
   $args{'site-defaults-file'}		= $args{'site-data-dir'}.'defaults';
   $args{'site-default-unitinfo-file'}  = $args{'site-data-dir'}.'unitinfo-txt.default';

   $args{demo}				= 0;
   
   $args{action} 			= '' 		if(!$args{action});
   $args{target} 			= 'home.plc' 	if(!$args{target});


# send http header and page content
$r->header_out(Pragma => 'no-cache');
$r->header_out('Cache-control' => 'no-cache');
$r->header_out(Connection => 'close');

$r->content_type('text/html');
$r->send_http_header;

# If the UserAgent only requested the HTTP header (using HEAD perhaps),
# we should send it and then exit, skipping the rest of the HTML document.

return 'OK' if $r->header_only;

my $sitename = 'Folding@Home Client Info';
	
htmlheader( title		 => $sitename,
	    link1       	 => 'stylesheet.css',
	    link1_rel   	 => 'stylesheet',
	    doctype              => 'HTML',
	    doctype_dtd          => 'HTML 4.01',
	    doctype_transitional => 1
	);

# check existance/permissions of direcotries used by 

my ($error, $msg) = (0, '');

my ($e1, $m1) = check_directory("$args{'client-data-dir'}", ('r', 'w'));
$error = 1   if($e1);
$msg  .= $m1 if($m1);

my ($e2, $m2) = check_directory("$args{'stanford-data-dir'}", ('r', 'w'));
$error = 1   if($e2);
$msg  .= $m2 if($m2);

my ($e3, $m3) = check_directory("$args{'bin-dir'}", ('r'));
$error = 1   if($e3);
$msg  .= $m3 if($m3);

my ($e4, $m4) = check_file("$args{'rasmol-bin'}", ('r', 'x'));
$error = 1   if($e4);
$msg  .= $m4 if($m4);

my ($e5, $m5) = check_file("$args{'xyz2pdb-bin'}", ('r', 'x'));
$error = 1   if($e5);
$msg  .= $m5 if($m5);

my ($e6, $m6) = check_directory("./", ('r', 'w'));
$error = 1   if($e6);
$msg  .= $m6 if($m6);

my ($e7, $m7) = check_directory("$args{'site-data-dir'}", ('r', 'w'));
$error = 1   if($e7);
$msg  .= $m7 if($m7);

my ($e8, $m8) = check_file("$args{'site-config-file'}", ('r', 'w'));
$error = 1   if($e8);
$msg  .= $m8 if($m8);

my ($e9, $m9) = check_directory("$args{'img-dir'}", ('r', 'w'));
$error = 1   if($e9);
$msg  .= $m9 if($m9);

my ($e10, $m10) = check_directory("$args{'qd-files-dir'}", ('r', 'w'));
$error = 1    if($e10);
$msg  .= $m10 if($m10);

my ($e11, $m11) = check_file("$args{'site-default-unitinfo-file'}", ('r'));
$error = 1    if($e11);
$msg  .= $m11 if($m11);

if($error) {
	print "<pre>\nPermission Check Failed, see INSTALL file for required permissions\n\nErrors:\n$msg</pre>\n";
	htmlfooter();
	exit;
}

# override global variable with user defined configuration
my %siteconfig = parse_siteconfig($args{'site-config-file'}, $args{'site-defaults-file'});

if($siteconfig{convert_path}) {
	my ($e0, $m0) = check_file("$args{'convert-bin'}", ('r', 'x'));
	$error = 1   if($e0);
	$msg  .= $m0 if($m0);

	if($error) {
		print "<pre>\nPermission Check Failed, see INSTALL file for required permissions\n\nErrors:\n$msg</pre>\n";
		htmlfooter();
		exit;
	}
	$args{'convert-bin'} = $siteconfig{convert_path};
}

$args{'user-summary-url'}     = $siteconfig{user_summary_url}     if($siteconfig{user_summary_url});
$args{'team-summary-url'}     = $siteconfig{team_summary_url}     if($siteconfig{team_summary_url});
$args{'project-summary-url'}  = $siteconfig{project_summary_url}  if($siteconfig{project_summary_url});


if($args{action} eq 'upload') {
	if($args{demo}) {
		print "Uploading not supported in demo!\n";
		htmlfooter();
		exit;
	}
	
	my $valid = 0;
	my $error = '';
	
	($valid, $error) = valid_upload(%args);
	if($valid != 1) {
		print "<pre>\nInvalid upload!\nErrors:\n$error\n</pre>\n";
		htmlfooter();
		exit;
	}

	print "<pre>\n";
	if( $siteconfig{auth_protect} ne 'upload' || 
	   ($siteconfig{auth_protect} eq 'upload' && $siteconfig{auth_username} eq $args{username} && $siteconfig{auth_password} eq $args{password}) ) {
		if($args{config_file} && $args{client_name} ne '') {
			# save client name in list file of all known client names
			my $updated = update_client_list($args{'client-list-file'}, $args{client_name});
			print "Successfully added client to client-list\n" if($updated); 

		
			# Write general info file for this client
			my $info_file = $args{"client-data-dir"}.'generalinfo.'.$args{client_name};
		
			if(!-e $info_file || -w $info_file) {
				open(F, ">$info_file") || die "Cannot open file ($!)";
				print F "last_update:$args{Date} $args{Time}\n";
				print F "client_ip:$args{RemoteIP}\n";
				if($args{client_version}) {
					print F "client_version:$args{client_version}\n";
				}
				if($args{qd_version}) {
					print F "qd_version:$args{qd_version}\n";
				}
				if($args{qdinfo_version}) {
					print F "qdinfo_version:$args{qdinfo_version}\n";
				}
				close F;
				
				print "Successfully saved general info file for client ($args{client_name})\n";
			}
			else {
				print "Error: saving general info file for client ($args{client_name}): Cannot write file!\n";
			}
	
			
			# Save unitinfo.txt file for the client
			my $unitinfo_file = $args{"client-data-dir"}.'unitinfo.'.$args{client_name};
			
			if(!-e $unitinfo_file || -w $unitinfo_file) {
				my $data = get_default_unitinfo($args{'site-default-unitinfo-file'});
				
				open(F, ">$unitinfo_file") || die "Cannot open file ($!)";
				if(!$args{unitinfo_file} || $args{unitinfo_file} eq '' ||$args{unitinfo_file} =~ /^\s*$/) {
					print F $data;
					
					print "Empty or no unitinfo.txt file uploaded, ";
					print "saved with default values for client ($args{client_name})\n";
				}
				else {
					print F $args{unitinfo_file};
		
					print "Successfully saved unitinfo file for client ($args{client_name})\n";
				}
				close F;
			}
			else {
				print "Error: saving unitinfo file for client ($args{client_name}): Cannot write file!\n";
			}
	
					
			# Save queue.dat file for the client if uploaded
			if($args{qd_output}) {
				my $qd_file = $args{"client-data-dir"}.'qd-output.'.$args{client_name};
			
				if(!-e $qd_file || -w $qd_file) {
					open(F, ">$qd_file") || die "Cannot open file ($!)";
					print F $args{qd_output};
					close F;
	
					print "Successfully saved qd output for client ($args{client_name})\n";
				}
				else {
					print "Error: saving qd output for client ($args{client_name}): Cannot write file!\n";
				}
			}
			
	
			# Save client.cfg file for this client
			my $config_file = $args{"client-data-dir"}.'client-cfg.'.$args{client_name};
		
			if(!-e $config_file || -w $config_file) {
				open(F, ">$config_file") || die "Cannot open file ($!)";
				print F $args{config_file};
				close F;
				
				print "Successfully saved config file for client ($args{client_name})\n";
			}
			else {
				print "Error: saving config file for client ($args{client_name}): Cannot write file!\n";
			}
	
			
			# create an .jpg image from the uploaded xyz file
			if($args{xyz_file}) {
				# first:  convert xyz to pdb file with xyz2pdb
	
				my $xyz_file = $args{"client-data-dir"}.'xyz.'.$args{client_name};
				my $pdb_file = $args{"client-data-dir"}.'pdb.'.$args{client_name};
				
				if(!-e $xyz_file || -w $xyz_file) {
					open(F, "> $xyz_file") || die "Cannot open file ($!)";
					print F $args{xyz_file};
					close F;
	
					if(!-e $pdb_file || -w $pdb_file) {
						my $xyz2pdb = `"$args{'xyz2pdb-bin'}" -xyzfile "$xyz_file" -o "$pdb_file"`;
						
						# second: generate ppm image(s) from pdb with rasmol
						
						my $script = '';
						
						# still image (JPG)
						my $rasmol_file = $args{'bin-dir'}.'rasmol.script';
						if($siteconfig{image_type} && $siteconfig{image_type} eq 'animation') {
							# animated image (GIF)
							$rasmol_file = $args{'bin-dir'}.'rasmol-rotate.script';
						}
						
						if(-r $rasmol_file) {
							open(F, $rasmol_file) || die "Cannot open file ($!)";
							while(<F>) {
								$script .= $_;
							}	
							close F;
							
							my $cmdline  = '';
							my $filebase = '';
							if($siteconfig{image_type} && $siteconfig{image_type} eq 'animation') {
								# animated image (GIF)
								
								my $turn  = 6;
								   $turn  = 360 / $siteconfig{image_frames} if($siteconfig{image_frames} && 
								   					       $siteconfig{image_frames} > 0);
								my $delay = 10;
								   $delay = 0                               if($siteconfig{image_delay} eq '0.0');
								   $delay = 10 * $siteconfig{image_delay}   if($siteconfig{image_delay} &&
								   					       $siteconfig{image_delay} > 0);
								my $axis  = 'y';
								   $axis  = $siteconfig{image_rotationaxis} if($siteconfig{image_rotationaxis} &&
								   					       $siteconfig{image_rotationaxis} =~ /^(x|y|z)$/);
								   
								my $commands = '';
								my $round    = 0;
								my $i        = 0;
								while($round < 360) {
									my $file = $args{'client-data-dir'}.'frame'.$i.'.'.$args{client_name}.'.ppm';

									$commands .= "rotate $axis $turn\n";
									$commands .= "write ppm $file\n";
									
									$cmdline  .= " -delay $delay" if($delay > 0);
									$cmdline  .= " $file";
									
									$round += $turn;
									$i++;
								}
								
								$script =~ s/COMMANDS\n/$commands/;
							}
							else {
								# still image (JPG)
							
								$filebase = $args{'client-data-dir'}.'protein.'.$args{client_name};
								$script =~ s/image/$filebase/;
							}
							
							my $rasmol_client_file = $args{"client-data-dir"}.'rasmol-script.'.$args{client_name};
							if(!-e $rasmol_client_file || -w $rasmol_client_file) {
								open(F, "> $rasmol_client_file") || die "Cannot open file ($!)";
								print F $script;
								close F;
								
								my $rasmol = `"$args{'rasmol-bin'}" -nodisplay "$pdb_file" < "$rasmol_client_file"`;
								
								if($siteconfig{image_type} && $siteconfig{image_type} eq 'animation') {
									# animated image (GIF)

									# create animated gif from rasmol's generated images
									my $gif_img = $args{"img-dir"}.'protein.'.$args{client_name}.'.gif';
									if(!-e $gif_img || -w $gif_img) {
										my $size = '200x200';
										   $size = $siteconfig{image_size} if($siteconfig{image_size});
										
										my $convert = `"$args{'convert-bin'}" $cmdline -geometry $size $gif_img`;
										
										print "Successfully created image of current protein\n";
									}
									else {
										print "Error: failed to create image!\n";
									}
								}
								else {
									# still image (JPG)
									
									# save as jpg resized
									my $ppm_img = $filebase.'.ppm';
									my $jpg_img = $args{"img-dir"}.'protein.'.$args{client_name}.'.jpg';
			
									if(resize_image($ppm_img, $jpg_img, 300, 300)) {
										print "Successfully saved image of current protein\n";
									}
									else {
										print "Error: failed to resize image!\n";
									}
								}
							}
							else {
								print "Error: saving rasmol client file for client ($args{client_name}): Cannot write file!\n";
							}
						}
						else {
							print "Error: reading rasmol file!\n";
						}
					}
					else {
						print "Error: saving pdb file for client ($args{client_name}): Cannot write file!\n";
					}
				}
				else {
					print "Error: saving xyz file for client ($args{client_name}): Cannot write file!\n";
				}
			}
	
	
			# save project name in list file with all currently active projects
			   $updated  = update_project_list($args{'project-list-file'}, $args{'client-list-file'}, $args{'client-data-dir'}, $args{'project-summary-file'});
			print "Successfully added project to project-list\n" if($updated);
	
			
			# save username & client name in list file
			   $updated  = update_username_list($args{'username-list-file'}, $args{'client-list-file'}, $args{'client-data-dir'}, $args{'project-summary-file'});
			print "Successfully added username to username-list\n" if($updated);
		}
		else {
			if(!$args{unitinfo_file} || $args{unitinfo_file} eq '') {
				print "ERROR: No unitinfo file uploaded!\n";
			}
			elsif(!$args{config_file} || $args{config_file} eq '') {
				print "ERROR: No config file uploaded!\n";
			}
			elsif(!$args{client_name} || $args{client_name} eq '') {
				print "ERROR: No client name specified!\n";
			}
			else {
				print "ERROR: Fatal Exception!\n";
			}
		}
	}
	else {
		if($siteconfig{auth_protect} eq 'upload') {
			print "Invalid username and/or password!\n";
		}
	}
	print "</pre>\n";
}
else {
	print "<!-- start main -->\n";
	print "<table align='center' width='95%' cellspacing='0' class='table-border'>\n";
	print "<tr><th colspan='2' class='table-head'>$sitename</th></tr>\n";

	print "<tr><td valign='top' class='right-border' width='20%'>\n\n";

	call('menu.plc', %args);	

	print "\n</td><td valign='top'>\n\n";
	
	call($args{target},%args);
	
	print "\n</td></tr>\n";
	print "</table>\n";

	print "<!-- end main -->\n";
}

htmlfooter();


#######################[ Sub Routines ]############################

sub iso_date {
       my $date     = $_[0];
       my $iso_date = $date;

       # Tue Jan 17 13:37:47 2006
       if($date =~ /^(\w+)\s+(\w+)\s+(\d+)\s+(\d+:\d+:\d+)\s+(\d+)$/) {
               my $dow   = $1;
               my $month = $2;
               my $day   = $3;
               my $time  = $4;
               my $year  = $5;

               $month = Date::Calc::Decode_Month($month);

               $day   = '0'.$day   if(length($day)   == 1);
               $month = '0'.$month if(length($month) == 1);

               $iso_date = "$year-$month-$day $time";
       }

       return $iso_date;
}

sub resize_image {
        my $source_file  = $_[0];
        my $target_file  = $_[1];
	my $width  	 = $_[2];
	my $height 	 = $_[3];
	my $blur   	 = 1;
	
	if(-r $source_file && (!-e $target_file || -w $target_file)) {
		my $image = Image::Magick->new(magick => 'jpg');

		$image->Read($source_file);

		my ($img_width, $img_height) = $image->Get('width','height');

		if($img_width > $img_height) {
			# Landscape orientated image

			if($img_width > $width) {
				# image is bigger that target size, so resize it
				
				my $new_width  = $width;
				my $new_height = $img_height / ($img_width / $new_width);

				$image->Resize(
					height => $new_height,
					width  => $new_width,
					blur   => $blur
				);
			}
		}
		elsif($img_height > $img_width) {
			# portrait orientated image
			
			if($img_height > $height) {
				# image is bigger that target size, so resize it

				my $new_height = $height;
				my $new_width  = $img_width / ($img_height / $new_height);

				$image->Resize(
					height => $new_height,
					width  => $new_width,
					blur   => $blur
				);
			}
		}
		else {
			# square image (width eq heigth)
			
			if($img_height > $height) {
				# image is bigger that target size, so resize it

				my $new_height = $height;
				my $new_width  = $img_width / ($img_height / $new_height);

				$image->Resize(
					height => $new_height,
					width  => $new_width,
					blur   => $blur
				);
			}
			elsif($img_width > $width) {
				# image is bigger that target size, so resize it
				
				 my $new_width  = $width;
				 my $new_height = $img_height / ($img_width / $new_width);

				$image->Resize(
					height => $new_height,
					width  => $new_width,
					blur   => $blur
				);
			}
			else {
				# image is bigger that target size, so resize it

				my $new_width  = $width;
				my $new_height = $img_height / ($img_width / $new_width);
				
				$image->Resize(
					height => $new_height,
					width  => $new_width,
					blur   => $blur
				);
			}
		}
		
		# write resized file to disc
		$image->Write( filename => $target_file );
		
		return 1;
	}
	else {
		return undef;
	}
}

sub user_rank {
	my $file = $_[0];
	my $user = $_[1];
	my $team = $_[2];
	
	my %info;
	if(-r $file) {
		my $i = 0;
		my %team;
		my %overall;

		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;
			
			#Fri Feb 18 12:00:00 PST 2005
			#name           round(credit)   total   team
			#anonymous      18706376        468476  0
			
			if($i >= 2 && (my @field = split /\t/)) {
				if($field[3] && $field[3] eq $team) {
					$team{$field[0]} = $field[1];
				}
				$overall{'team'.$field[3].'user'.$field[0]} = $field[1];
			}
			
			$i++;
		}
		close F;
		
		$i = 1;
		foreach(sort { $team{$b} <=> $team{$a} } keys %team) {
			if($_ eq $user) {
				$info{team} = $i;
				last;
			}
			$i++;
		}

		$i = 1;
		foreach(sort { $overall{$b} <=> $overall{$a} } keys %overall) {
			if($_ eq 'team'.$team.'user'.$user) {
				$info{overall} = $i;
				last;
			}
			$i++;
		}
	}

	return %info;
}

sub parse_siteconfig {
	my $config_file   = $_[0];
	my $defaults_file = $_[1];

	my %info;
	my %default;
	if(-r $config_file) {
		open(F, $config_file) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;

			if(/^(.*?)=(.*)$/) {
				$info{$1} = $2;
			}
		}
		close F;
	}
	if(-r $defaults_file) {
		open(F, $defaults_file) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;
			
			if(/^(.*?)=(.*)$/) {
				$info{$1} = $2 if(!$info{$1});
			}
		}
		close F;
	}
	
	return %info;
}
		
sub parse_lastupdate {
	my $file = $_[0];

	my $update = 'n/a';
	if(-r $file) {
		open(F, $file) || die "Cannot open file ($!)";
		chomp($update = <F>);
		close F;
	}

	return $update;
}

sub parse_usersummary {
	my $file     = $_[0];
	my $username = $_[1] ? $_[1] : '';

	my %info;
	if(-r $file) {
		my $i = 0;

		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;
			
			if(s/\r//g) {}

			#Thu Jan 13 03:00:01 PST 2005
			#name    round(credit)   total   team
			#anonymous       15743946        437183  0

			if($i >= 2 && $username eq '') {
				my ($name, $credit, $total, $team) = split /\t/;

				$info{$team}{$name}{name}   = $name;
				$info{$team}{$name}{credit} = $credit;
				$info{$team}{$name}{total}  = $total;
				$info{$team}{$name}{team}   = $team;
			}
			elsif($i >= 2 && ($username ne '' && /^\Q$username\E\t/) ) {
				my ($name, $credit, $total, $team) = split /\t/;

				$info{$team}{name}   = $name;
				$info{$team}{credit} = $credit;
				$info{$team}{total}  = $total;
				$info{$team}{team}   = $team;
			}

			$i++;
		}
		close F;
	}

	return %info;
}

sub parse_teamsummary {
	my $file = $_[0];
	my $team = $_[1] ? $_[1] : '';

	my %info;
	if(-r $file) {
		my $i = 0;

		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;
			
			if(s/\r//g) {}
			
			#Thu Jan 13 03:00:01 PST 2005
			#team    teamname        score   wu
			#0       Default (includes all those WU returned without valid team number)      231453370.603365        6123294
			#446     Google  220443432.95776 4175421

			if($i >= 2 && $team eq '') {
				my ($team, $teamname, $score, $wu) = split /\t/;

				$info{$team}{team}  = $team;
				$info{$team}{name}  = $teamname;
				$info{$team}{score} = $score;
				$info{$team}{wu}    = $wu;
			}
			elsif($i >= 2 && ($team ne '' && /^\Q$team\E\t/) ) {
				my ($team, $teamname, $score, $wu) = split /\t/;
				
				$info{team}  = $team;
				$info{name}  = $teamname;
				$info{score} = $score;
				$info{wu}    = $wu;
			}

			$i++;
		}
		close F;
	}

	return %info;
}

sub parse_projectlist {
	my $file = $_[0];
	
	my %info;
	if(-r $file) {
		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;
	
			if(s/\r//g) {}
			if(/^(.*):(.*?)$/) {
				push @{$info{$1}}, $2;
			}
		}
		close F;
	}

	return %info;
}

sub parse_usernamelist {
	my $file = $_[0];

	my %info;
	if(-r $file) {
		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;
			
			if(s/\r//g) {}
			if(/^(.*):(.*?):(.*?)$/) {
				my ($username, $team, $client) = ($1, $2, $3);
			
				push @{$info{$username}{$team}}, $client;
			}
			else {
				print "Error: failed parsing line: $_\n" unless(/^#/);
			}
		}
		close F;
	}

	return %info;
}

sub parse_configinfo {
	my $file = $_[0];

	my %info;
	if(-r $file) {
		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;
			
			if(s/\r//g) {}
			if(/^(.*?)=(.*)$/) {
				$info{$1} = $2;
			}
		}
		close F;
	}

	return %info;
}

sub parse_generalinfo {
	my $file = $_[0];

	my %info;
	if(-r $file) {
		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;
			
			if(s/\r//g) {}
			if(/^(.*?):(.*)$/) {
				$info{$1} = $2;
			}
		}
		close F;
	}

	return %info;
}

sub parse_unitinfo {
	my $file = $_[0];
	
	my %info;
	if(-r $file) {
		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;

			if(s/\r//g) {}
			
			#Name: p1134_RIBO_FSpeptide_EXT
			if(/^(.*?): (.*)$/) {
				$info{lc($1)} = $2;
			}
		}
		close F;
	}

	return %info;
}

sub parse_psummary {
        my $file = $_[0];
	my $name = $_[1] ? $_[1] : '';
	
	my %info;
	if(-r $file) {
		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			# Example line to parse
			#<TR bgcolor="#dfdfdf"> <TD>212</TD><TD> 171.64.122.124 </TD><TD>p212_villinURE212</TD><TD>20605</TD><TD>   27.00</TD>
			#<TD>   39.00</TD><TD>  135.00</TD><TD>100</TD><TD>GROMACS</TD>
			#<TD><a href=http://VSPX27.stanford.edu/cgi-bin/allprojects#212>Description</a></TD>
			#<TD><font size=-1>rbrandman</font></TD></TR>
			
			if(m!<TR bgcolor="#\w{6}">\s*<TD>(\d+)</TD><TD>\s*(\d+\.\d+\.\d+\.\d+)\s*</TD><TD>(.*?)</TD><TD>(\d+)</TD><TD>\s*(\d+\.\d+)</TD><TD>\s*(.*?)</TD><TD>\s*(\d+\.\d+)</TD><TD>(\d+)</TD><TD>(.*?)</TD><TD><a href=(.*?)>Description</a></TD><TD><font size=-1>(.*?)</font></TD></TR>!) {
			
				my ($project_number, $server_ip, $wu_name, $number_of_atoms, $preferred_days, $final_deadline,
				    $credit, $frames, $code, $description_url, $contact) = ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11);
				
				if($name eq '') {
					$info{$wu_name}{project_number}  = $project_number;
					$info{$wu_name}{server_ip}       = $server_ip;
					$info{$wu_name}{wu_name}         = $wu_name;
					$info{$wu_name}{number_of_atoms} = $number_of_atoms;
					$info{$wu_name}{preferred_days}  = $preferred_days;
					$info{$wu_name}{final_deadline}  = $final_deadline;
					$info{$wu_name}{credit}          = $credit;
					$info{$wu_name}{frames}          = $frames;
					$info{$wu_name}{code}            = $code;
					$info{$wu_name}{description_url} = $description_url;
					$info{$wu_name}{contact}         = $contact;
				}
				elsif($name ne '' && $name =~ /^p(\d+)_/ && $wu_name =~ /^$name/) {
					$info{project_number}  = $project_number;
					$info{server_ip}       = $server_ip;
					$info{wu_name}         = $wu_name;
					$info{number_of_atoms} = $number_of_atoms;
					$info{preferred_days}  = $preferred_days;
					$info{final_deadline}  = $final_deadline;
					$info{credit}          = $credit;
					$info{frames}          = $frames;
					$info{code}            = $code;
					$info{description_url} = $description_url;
					$info{contact}         = $contact;
				}
				elsif($name ne '' && $name !~ /^p(\d+)_/) {
					$info{project_number}  = 'n/a';
					$info{server_ip}       = 'n/a';
					$info{wu_name}         = 'n/a';
					$info{number_of_atoms} = 'n/a';
					$info{preferred_days}  = 'n/a';
					$info{final_deadline}  = 'n/a';
					$info{credit}          = 'n/a';
					$info{frames}          = 'n/a';
					$info{code}            = 'n/a';
					$info{description_url} = 'n/a';
					$info{contact}         = 'n/a';
				}
				
			}
			else {
				#print "ERROR Parsing line: $_";
			}
		}
		close F;
	}
	
	return %info;
}

sub parse_qdoutput {
	my $file = $_[0];
	
	my %info;
	if(-r $file) {
		my $key = '';
		
		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			#qd released 6 February 2005 (fr 028)
			if(/qd released (.*? \(.*?\))/) {
				$info{qd_release} = $1;
			}
			
			#qd info 6 February 2005 (rph)
			if(/qd info (.*)/) {
				$info{qdinfo_release} = $1;
			}
				
			#Queue version 5.01
			if(/Queue version (\d+\.\d+.*?)/) {
				$info{queue_version} = $1;
			}
			
			#Current index: 9
			if(/Current index: (\d+)$/) {
				$info{current_index} = $1;
			}
			
			# Index 0: finished 302.00 pts (8.529 pt/hr) 22.4 X min speed
			#  server: 171.65.103.154:8080; project: 1322
			#  Folding: run 0, clone 494, generation 1; benchmark 1384; misc: 500, 400
			#  issue: Fri Jan  7 03:10:50 2005; begin: Fri Jan  7 03:25:03 2005
			#  end: Sat Jan  8 14:49:35 2005; due: Wed Feb  9 03:25:03 2005 (33 days)
			#  core URL: http://www.stanford.edu/~pande/Linux/x86/Core_78.fah
			#  assignment info (le): Fri Jan  7 03:10:49 2005; A2288C4C
			#  CS: 171.65.103.100; P limit: 52427776
			#  user: [DPC]_Fatal_Error_Group0smoking2000; team: 92; ID: 637FFF483558C82F; mach ID: 1
			#  work/wudata_00.dat file size: 862570; WU type: Folding@Home
			
			# Index 9: folding now
			# Index 0: finished
			if(/Index (\d+): (empty|deleted|finished|garbage|folding now|queued for processing|ready for upload|DANGER will be lost if client is restarted!|abandoned|fetching from server|UNKNOWN STATUS = \S+)/) {

				$key = 'index_'.$1;
					
				$info{$key}{wu_status} = $2;
			}
			
			#302.00 pts
			if(/(\d+(\.\d+)?) pts/) {
				$info{$key}{wu_score} = $1;
			}

			#(8.529 pt/hr)
			if(/\((\d+(\.\d+)?) pt\/hr\)/) {
				$info{$key}{score_per_hour} = $1;
			}

			#7.23 X min speed
			if(/(\d+(\.\d+)?) X min speed/) {
				$info{$key}{progress_rate} = $1;
			}

			#69% complete
			if(/(\d+(\.\d+)?%) complete/) {
				$info{$key}{percent_complete} = $1;
			}

			#server: 171.65.103.154:8080
			if(/server: (\d+\.\d+\.\d+\.\d+(:\d+)?)/) {
				$info{$key}{server} = $1;
			}
			#server 1: 171.65.103.154
			#server 2: 171.65.103.154:8080
			if(/server (\d+): (\d+\.\d+\.\d+\.\d+(:\d+)?)/) {
				my $server = 'server'.$1;
				
				$info{$key}{$server} = $2;
			}

			#project: 1322, "p1475_tet1_03_1"
			if(/project: (\d+), "(.*?)"/) {
				$info{$key}{project} = $1;
				$info{$key}{wu_name} = $2;
			}
			#project: 1322
			elsif(/project: (\d+)/) {
				$info{$key}{project} = $1;
			}
				
			#Genome: unit 36476, 7554, 656
			if(/Genome: unit (\d+, \d+, \d+)/) {
				$info{$key}{genome_unit} = $1;
			}
			
			#Folding: run 0, clone 494, generation 1
			if(/Folding: run (\d+), clone (\d+), generation (\d+)/) {
				$info{$key}{folding_run}        = $1;
				$info{$key}{folding_clone}      = $2;
				$info{$key}{folding_generation} = $3;
			}

			#benchmark 1384
			if(/benchmark (\d+)/) {
				$info{$key}{benchmark} = $1;
			}

			#misc: 500, 400
			#misc: 500, 400; 500, 400
			if(/misc: (\d+, \d+(; \d+, \d+)?)/) {
				$info{$key}{misc} = $1;
			}

			#issue: Fri Jan 28 10:06:08 2005
			#issue: Fri Jan  7 03:10:50 2005
			if(/issue: (\w+\s+\w+\s+\d+\s+\d+:\d+:\d+\s+\d+)/) {
				$info{$key}{issue} = $1;
			}
			
			#begin: Fri Jan  7 03:25:03 2005
			if(/begin: (\w+\s+\w+\s+\d+\s+\d+:\d+:\d+\s+\d+)/) {
				$info{$key}{begin} = $1;
			}
				
			#end: Tue Feb  1 20:56:19 2005
			#end: Tue Feb  1 20:56:19 2005
			if(/end: (\w+\s+\w+\s+\d+\s+\d+:\d+:\d+\s+\d+)/) {
				$info{$key}{end_date} = $1;
			}
			
			#due: Wed Apr 20 11:21:54 2005 (82 days)
			#due: Wed Apr 20 11:21:54 2005 (10 hours)
			if(/due: (\w+\s+\w+\s+\d+\s+\d+:\d+:\d+\s+\d+) \((.*?) (days|hours)\)/) {
				$info{$key}{due_date} = $1;
				$info{$key}{due_time} = "$2 $3";
			}
			
			#expire: Wed Apr 20 11:21:54 2005 (82 days)
			#expire: Wed Apr 20 11:21:54 2005 (10 hours)
			if(/expire: (\w+\s+\w+\s+\d+\s+\d+:\d+:\d+\s+\d+) \((.*?) (days|hours)\)/) {
				$info{$key}{expire_date} = $1;
				$info{$key}{expire_time} = "$2 $3";
			}
			
			#expect: Wed Feb  9 01:33:16 2005
			if(/expect: (\w+\s+\w+\s+\d+\s+\d+:\d+:\d+\s+\d+)/) {
				$info{$key}{expect_date} = $1;
			}
			
			#core URL: http://www.stanford.edu/~pande/Linux/x86/Core_78.fah (V1.70)
			if(/core URL: (.*) \((V.*?)\)$/) {
				$info{$key}{core_url}     = $1;
				$info{$key}{core_version} = $2;
			}
			#core URL: http://www.stanford.edu/~pande/Linux/x86/Core_78.fah
			#core URL: MISSING
			elsif(/core URL: (.*)/) {
				$info{$key}{core_url} = $1;
			}

			#CPU: 1,0 x86
			#CPU: 1,2000 AMD x86
			#CPU: 1,87 Pentium IV
			if(/CPU: (\d+,\d+) (.*);/) {
				$info{$key}{cpu_type} = $1;
				$info{$key}{cpu_name} = $2;
			}
			#CPU: 1,87
			elsif(/CPU: (\d+,\d+)/) {
				$info{$key}{cpu_type} = $1;
			}

			#OS: 4,0 Linux
			#OS: 1,8 WinXP
			#OS: 1,9 Win2K3
			if(/OS: (\d+,\d+) (.*)/) {
				$info{$key}{os_type} = $1;
				$info{$key}{os_name} = $2;
			}
			#OS: 1,9
			elsif(/OS: (\d+,\d+)/) {
				$info{$key}{os_type} = $1;
			}

			#assignment info (le): Fri Jan  7 03:10:49 2005; A2288C4C
			#assignment info (be): Fri Jan  7 03:10:49 2005; A2288C4C
			if(/assignment info \((le|be)\): (.*?); (.*)/) {
				$info{$key}{assinfo_endian}    = $1;
				$info{$key}{assinfo_timestamp} = $2;
				$info{$key}{assinfo_checksum}  = $3;
			}

			#CS: 171.67.89.100
			if(/CS: (\d+\.\d+\.\d+\.\d+)/) {
				$info{$key}{collection_server} = $1;
			}

			#DL: Sat Sep 18 01:47:09 2004
			if(/DL: (\w+\s+\w+\s+\d+\s+\d+:\d+:\d+\s+\d+)/) {
				$info{$key}{work_downloaded} = $1;
			}

			#upload failures: 1
			if(/upload failures: (\d+)/) {
				$info{$key}{upload_failures} = $1;
			}

			#P limit: 52427776
			if(/P limit: (\d+)/) {
				$info{$key}{packet_size_limit} = $1;				
			}

			#user: [DPC]_Fatal_Error_Group0smoking2000; team: 92; ID: 637FFF483558C82F; mach ID: 1
			if(/user: (.*); team: (\d+); ID: (.*?); mach ID: (\d+)/) {
				$info{$key}{username}  = $1;
				$info{$key}{team}      = $2;
				$info{$key}{id}        = $3;
				$info{$key}{machineid} = $4;
			}
			#user: [DPC]_Fatal_Error_Group0smoking2000; team: 92; ID: 637FFF483558C82F
			elsif(/user: (.*); team: (\d+); ID: (.*?)/) {
				$info{$key}{username}  = $1;
				$info{$key}{team}      = $2;
				$info{$key}{id}        = $3;
			}
			#user: [DPC]_Fatal_Error_Group0smoking2000; ID: 637FFF483558C82F; mach ID: 1
			elsif(/user: (.*); ID: (.*?); mach ID: (\d+)/) {
				$info{$key}{username}  = $1;
				$info{$key}{id}        = $2;
				$info{$key}{machineid} = $3;
			}
			#user: [DPC]_Fatal_Error_Group0smoking2000; ID: 637FFF483558C82F
			elsif(/user: (.*); ID: (.*?)/) {
				$info{$key}{username}  = $1;
				$info{$key}{id}        = $2;
			}
				
			#work/wudata_00.dat file size: 862570
			if(/(work\/wudata_\d+.dat) file size: (\d+)/) {
				$info{$key}{wudata_filename} = $1;
				$info{$key}{wudata_filesize} = $2;
			}
				
			#WU type: Folding@Home
			if(/WU type: (Folding\@Home|.*?)/) {
				$info{$key}{wutype} = $1;
			}

			#Average download rate 175.261 KB/s (u=4)
			if(/Average download rate (\d+\.\d+ KB\/s) \(u=(\d+)\)/) {
				$info{download_rate}  = $1;
				$info{download_units} = $2;
			}

			#upload rate 34.042 KB/s (u=4)
			if(/upload rate (\d+\.\d+ KB\/s) \(u=(\d+)\)/) {
				$info{upload_rate}  = $1;
				$info{upload_units} = $2;
			}

			#Performance fraction 0.954537 (u=4)
			if(/Performance fraction (\d+\.\d+) \(u=(\d+)\)/) {
				$info{performance_fraction} = $1;
				$info{performance_units}    = $2;
			}
			
			#Average pph: 6.194, ppd: 148.65, ppw: 1040.5, ppy: 54292
			if(/Average pph: (\d+\.\d+), ppd: (\d+\.\d+), ppw: (\d+\.\d+), ppy: (\d+)/) {
				$info{average_pph} = $1;
				$info{average_ppd} = $2;
				$info{average_ppw} = $3;
				$info{average_ppy} = $4;
			}

			#print "<pre>$_</pre>\n";
		}
		close F;
	}
	
	return %info;
}

sub valid_upload {
	my %args  = @_;
	my $valid = 1;
	my $error = '';

	if(!$args{client_name} || $args{client_name} =~ /^\s*$/) {
		$valid  = 0;
		$error .= "Clientname cannot be blank!\n";
	}
	elsif($args{client_name} !~ /^[0-9a-zA-Z\-\_]+$/) {
		$valid  = 0;
		$error .= "Clientname contains illegal characters: alphanumic, score and underscore only!\n";
	}

	return ($valid, $error);
}

sub save_file {
	my $filename = $_[0];
	my $content  = $_[1];
	
	if(!-e $filename || -w $filename) {
		open(F, "> $filename") || die "Cannot open file ($!)";
		print F $content;
		close F;
		
		return 1;
	}
	else {
		return undef;
	}
}

sub fetch_url {
	my $url = $_[0];
	
	my $ua  = LWP::UserAgent->new();
	my $req = HTTP::Request->new(GET => $url);
	my $res = $ua->request($req);
	
	if($res->is_success) {
		return $res->content;
	}
	else {
		return undef;
	}
}

sub valid_url {
	my $url = $_[0];
	
	my $ua  = LWP::UserAgent->new();
	my $req = HTTP::Request->new(HEAD => $url);
	my $res = $ua->request($req);
	
	if($res->is_success()) {
		return 1;
	}
	elsif($res->is_error()) {
		return 0;
	}
	else {
		return 0;
	}
}
																	   

sub check_file {
	my ($file, @perm) = @_;
	my $error = 0;
	my $msg = '';

	foreach(@perm) {
		if($_ eq 'r') {
			if(!-r $file) {
				$error = 1;
				$msg  .= "Cannot read file! ($file)\n";
			}
		}
		elsif($_ eq 'w') {
			if(!-w $file) {
				$error = 1;
				$msg  .= "Cannot write file! ($file)\n";
			}
		}
		elsif($_ eq 'x') {
			if(!-x $file) {
				$error = 1;
				$msg  .= "Cannot execute file! ($file)\n";
			}
		}
	}

	return ($error, $msg);
}

sub check_directory {
	my ($dir, @perm) = @_;
	my $error = 0;
	my $msg = '';

	foreach(@perm) {
		if($_ eq 'r') {
			if(!-r $dir) {
				$error = 1;
				$msg  .= "Cannot read directory! ($dir)\n";
			}
		}
		elsif($_ eq 'w') {
			if(!-w $dir) {
				$error = 1;
				$msg  .= "Cannot write directory! ($dir)\n";
			}
		}
	}

	return ($error, $msg);
}

sub update_client_list {
	my $file       = $_[0];
	my $client     = $_[1];
	my $new_client = 1;

	if(-r $file && -w $file) {
		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			if(/^\Q$client\E$/) {
				$new_client = 0;
			}
		}
		close F;

		if($new_client == 1) {
			open(F, ">> $file") || die "Cannot open file ($!)";
			print F "$client\n";
			close F;
		}
	}
	else {
		if(!-e $file) {
			# file doesn't exists, create it

			if($new_client == 1) {
				open(F, ">> $file") || die "Cannot open file ($!)";
				print F "$client\n";
				close F;
			}
		}
		else {
			# file does exists, debug
			
			if(!-r $file) {
				print "Error: Cannot read file ($file) !\n";
				return -1;
			}
			elsif(!-w $file) {
				print "Error: Cannot write file ($file) !\n";
				return -1;
			}
			else {
				print "Error: Fatal Exeption on file ($file) !\n";
				return -1;
			}
		}
	}

	return $new_client;
}

sub update_username_list {
	my $output_file     = $_[0];	#$args{'username-list-file'}
	my $client_list     = $_[1];	#$args{'client-list-file'}
	my $client_data_dir = $_[2];	#$args{'client-data-dir'}
	my $success         = 1;
	
	if(-r $output_file && -w $output_file) {
		$success = generate_username_list($output_file, $client_list, $client_data_dir);
	}
	else {
		if(!-e $output_file) {
			# file doesn't exists, create it
			
			$success = generate_username_list($output_file, $client_list, $client_data_dir);
		}
		else {
			# file does exists, debug
			
			if(!-r $output_file) {
				print "Error: Cannot read file ($output_file) !\n";
				return -1;
			}
			elsif(!-w $output_file) {
				print "Error: Cannot write file ($output_file) !\n";
				return -1;
			}
			else {
				print "Error: Fatal Exeption on file ($output_file) !\n";
				return -1;
			}
		}
	}

	return $success;
}

sub generate_username_list {
	my $output_file      = $_[0];
	my $client_list_file = $_[1];	#$args{'client-list-file'}
	my $client_data_dir  = $_[2];	#$args{'client-data-dir'}
	
	my @client;
	open(F, $client_list_file) || die "Cannot open file ($!)";
	while(<F>) {
		chomp;
		
		push @client, $_;
	}
	close F;

	my $data = '';
	foreach(@client) {
		my $client = $_;

		my $config_file = $client_data_dir.'client-cfg.'.$client;
		my %configinfo  = parse_configinfo($config_file);

		$data .= "$configinfo{username}:$configinfo{team}:$client\n";
	}

	open(F, "> $output_file") || die "Cannot open file ($!)";
	print F $data;
	close F;

	return 1;
}

sub update_project_list {
	my $output_file      = $_[0];
	my $client_list_file = $_[1];   #$args{'client-list-file'}
	my $client_data_dir  = $_[2];   #$args{'client-data-dir'}
	my $psummary_file    = $_[3];	#$args{'project-summary-file'}
	
	my $success = 1;
	
	if(-r $output_file && -w $output_file) {
		$success = generate_project_list($output_file, $client_list_file, $client_data_dir, $psummary_file);
	}
	else {
		if(!-e $output_file) {
			# file doesn't exists, create it
			
			$success = generate_project_list($output_file, $client_list_file, $client_data_dir, $psummary_file);
		}
		else {
			# file does exists, debug
			
			if(!-r $output_file) {
				print "Error: Cannot read file ($output_file) !\n";
				return -1;
			}
			elsif(!-w $output_file) {
				print "Error: Cannot write file ($output_file) !\n";
				return -1;
			}
			else {
				print "Error: Fatal Exeption on file ($output_file) !\n";
				return -1;
			}
		}
	}

	return $success;
}

sub generate_project_list {
	my $output_file      = $_[0];
	my $client_list_file = $_[1]; 	#$args{'client-list-file'}
	my $client_data_dir  = $_[2];	#$args{'client-data-dir'}
	my $psummary_file    = $_[3];	#$args{'project-summary-file'}
	
	my @client;
	open(F, $client_list_file) || die "Cannot open file ($!)";
	while(<F>) {
		chomp;
		
		push @client, $_;
	}
	close F;

	my $data = '';
	foreach(@client) {
		my $client = $_;

		my $unitinfo_file = $client_data_dir.'unitinfo.'.$client;
		my %unitinfo      = parse_unitinfo($unitinfo_file);
		my %wu_info	  = parse_psummary($psummary_file, $unitinfo{name});

		$wu_info{wu_name} = $unitinfo{name} if(!$wu_info{wu_name});

		$data .= "$wu_info{wu_name}:$client\n";
	}

	open(F, "> $output_file") || die "Cannot open file ($!)";
	print F $data;
	close F;

	return 1;
}

sub get_default_unitinfo {
	my $file = $_[0];

	my $data = '';
	if(-r $file) {
		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			$data .= $_;
		}
		close F;
		
		my ($year,$month,$day,$hour,$min,$sec) = Date::Calc::Today_and_Now();
	
		$month = Date::Calc::Month_to_Text($month);
		$hour  = '0'.$hour  if(length($hour)  == 1);
		$min   = '0'.$min   if(length($min)   == 1);
		$sec   = '0'.$sec   if(length($sec)   == 1);
		
		if($data =~ s/\#Month\#/$month/g) {}
		if($data =~ s/\#D\#/$day/g) {}
		if($data =~ s/\#HH\#/$hour/g) {}
		if($data =~ s/\#MM\#/$min/g) {}
		if($data =~ s/\#SS\#/$sec/g) {}
	}

	return $data;
}

#####[ Start helper subroutines for call() routine ]#####

sub call {
	# for security reseason, only allow scripts to be called in the current 
	# direcotory, the one which holds this index.pl script
	#
	# parse the script, show errors if any, and error messages are allowed 
	# to be displayed (debug/warning mode)
	#
	# eval() the parsed Perl code, show errors if any and warning mode is on
	#
	# execute the eval()'ed code, show errors if any and are allowed to be shown
	#
	# return any return value if any

	my ($file, %args) = @_;

	if($file =~ /\.plc$/ && $file !~ /\// && -r "./$file") {
		if(my $script = parse($file)) {
			my $code = eval $script;
			if ($@ || !$code) {
				print "<div style='color: #FF0000'>Error compiling $file <pre>$@</pre></div><br><hr>\n";
				return undef;
			}
			else {
				my $return;
				eval {
					$return = &$code(%args);
				};
				
				if($@) {
					print "<div style='color: #FF0000'>Error running $file: <pre>$@</pre></div><br><hr>\n";
				}
				else {
					return $return;
				}
			}
		}
		else {
			print "<div style='color: #FF0000'>Cannot load $file</div><br><hr>\n";
			return undef;
		}
	}
	else {
		print "<div style='color: #FF0000'>Cannot load $file</div><br>\n";
		return undef;
	}
}

sub parse {
	# compile all the Perl code sections from the requisted file into a 
	# single piece of code
	# 
	# return the code as a subroutine to be executed
	# return undef if no code is compiled
	
	my $file  = $_[0];
	my $parse = 0;
	my $data  = '';

	open(F, $file) || return undef;
	while(<F>) {
		if(/<\%perl>/i && !$parse) {
			$parse = 1;
			$data .= "\n";
		}
		elsif(/<\/\%perl>/i && $parse) {
			$parse = 0;
			$data .= "\n";
		}
		elsif(/<\%/ && !$parse) {
			while(/<%(.*?)%>/) {
				my $code    = $1;
				my $prefix  = $`;
				my $postfix = $';
				
				$prefix =~ s/'/'."'".'/g;
				$data  .= "print '$prefix';";
				$data  .= "print $code;";
				
				$_ = $postfix;
			}
			s/'/'."'".'/g;
			$data .= "print '$_';";
		}
		else {
			if($parse) {
				$data .= $_;
			}
			else {
				s/'/'."'".'/g;
				$data .= "print '$_';";
			}
		}
	}
	close F;
	
	if($data) {
		return 'return sub { my %args = @_; '.$data.' };';
	}
	else {
		return undef;
	}
}

#####[ End helper subroutines for call() routine ]#####

sub get_args {
	my ($r, $c, $s) = @_;
	my %args;

	# fill hash with URI parameters

	# handle POST requests
	if($r->method eq 'POST') {
		my $apr = Apache::Request->new($r);
		
		# normal parameters
		foreach my $param ($apr->param) {
			my $i = 0;
			foreach my $value ($apr->param($param)) {
				# parameters can have multiple values, 
				# number the ones afther the 1st

				$param .= $i if($i > 0);
				
				$args{$param} = $value;
		
				$i++;
			}
		}
		
		# uploaded files
		for my $upload ($apr->upload) {
			my $param = $upload->name;
			
			$args{$param.'_Size'}     = $upload->size;
			$args{$param.'_Type'}     = $upload->type;
			$args{$param.'_Tempname'} = $upload->tempname;
			$args{$param.'_Filename'} = $upload->filename;

			foreach(keys %{$upload->info}) {
				$args{$param.'_Info_'.$_} = $upload->info($_);
			}
			
			my $fh = $upload->fh;
			while (<$fh>) {
				$args{$param} .= $_;
			}
			
			# strip the filename from the start of the data
			if($args{$param} =~ s/^\Q$args{$param.'_Filename'}\E//) {}
			
			# handle files uploaded with Internet Explorer
			# instead of just the filename, IE sends the whole path
			# X:\Porn\Pics\NakedLady.jpg
			
			if($args{$param.'_Filename'} =~ /^\w:.*\\(.*?)$/) {
				$args{$param.'_Filename'} = $1;
			}
		}
	}
	# handle GET requests
	else {
		%args = $r->args;
	}
	
	my ($year,$month,$day,$hour,$min,$sec) = Date::Calc::Today_and_Now();
	
	$month = '0'.$month if(length($month) == 1);
	$day   = '0'.$day   if(length($day)   == 1);
	$hour  = '0'.$hour  if(length($hour)  == 1);
	$min   = '0'.$min   if(length($min)   == 1);
	$sec   = '0'.$sec   if(length($sec)   == 1);

	$args{'Date'}		= "$year-$month-$day";
	$args{'Time'}		= "$hour:$min:$sec";
	$args{'User'}           = $c->user		if defined($c->user);
	$args{'Protocol'}   	= $r->protocol;
	
	$args{'Bytes'}		= $r->bytes_sent	if($r->bytes_sent);
	$args{'Method'}		= $r->method;
	$args{'MethodNumber'}   = $r->method_number;
	$args{'Request'}        = $r->the_request;
	$args{'RequestTime'}    = $r->request_time;
	$args{'URI'}            = $r->uri;
	$args{'Filename'}       = $r->filename;
	$args{'PathInfo'}       = $r->path_info		if($r->path_info);
	$args{'DocumentRoot'}   = $r->document_root;
	
	$args{'UserAgent'}      = $r->header_in('User-Agent');	# for compatibility with old scripts	

	my %header = $r->headers_in;
	foreach(keys %header) {
		$args{$_} = $header{$_};
	}
	
	$args{'Hostname'}   	= $r->hostname;
	$args{'AuthName'}	= $r->auth_name		if($r->auth_name);
	$args{'AuthType'}	= $c->auth_type		if($c->auth_type);
	$args{'RemoteIP'}	= $c->remote_ip;
	if($c->remote_host() eq '') {
		$args{'RemoteHostname'} = $r->get_remote_host;
	}
	else {
		$args{'RemoteHostname'} = $c->remote_host;
	}

	$args{'ServerName'}     = $s->server_hostname;
	$args{'ServerAdmin'}    = $s->server_admin;
	$args{'ServerPort'}     = $s->port;
	$args{'ServerSoftware'} = $ENV{SERVER_SOFTWARE};	

	return %args;
}

sub htmlheader {
	my %args = @_;
	
	if($args{doctype}) {
		print "<!DOCTYPE $args{doctype} PUBLIC \"-//W3C//DTD ";
		if($args{doctype_dtd}) {
			print "$args{doctype_dtd} ";
		}
		else {
			print "$args{doctype} ";
		}
		
		if($args{doctype_transitional} == 1) {
			print "Transitional";
		}
		elsif($args{doctype_strict} == 1) {
			print "Strict";
		}
		elsif($args{doctype_loose} == 1) {
			print "Loose";
		}
		else {
			print "Transitional";
		}
		
		if($args{doctype_language}) {
			print "//$args{doctype_language}\">\n";
		}
		else {
			print "//EN\">\n";
		}
	}
	
	print "<html>\n";
	print "<head>\n";
	print "\t<title>$args{title}</title>\n" if($args{title});
	if($args{meta1}) {
		my $i = 1;
		while($args{'meta' . $i}) {
			print "\t<meta";
			
			my $meta_name = $args{'meta' . $i . '_name'} if($args{'meta' . $i . '_name'});
			print " name=\"$meta_name\"" if($args{'meta' . $i . '_name'});
			
			my $meta_equiv = $args{'meta' . $i . '_equiv'} if($args{'meta' . $i . '_equiv'});
			print " http-equiv=\"$meta_equiv\"" if($args{'meta' . $i . '_httpequiv'});
			
			my $content  = $args{'meta' . $i . '_content'} if($args{'meta' . $i . '_content'});
			print " content=\"$content\"" if($args{'meta' . $i . '_content'});
			
			$i++;
		}
	}
	if($args{link1}) {
		my $i = 1;
		while($args{'link' . $i}) {
			print "\t<link";
			
			my $linktype = $args{'link' . $i . '_type'} if($args{'link' . $i . '_type'});
			print " type=\"$linktype\"" if($args{'link' . $i . '_type'});
			
			my $linkrel  = $args{'link' . $i . '_rel'} if($args{'link' . $i . '_rel'});
			print " rel=\"$linkrel\"" if($args{'link' . $i . '_rel'});
			
			print " href=\"$args{'link' . $i}\">\n";
			
			$i++;
		}
	}
	if($args{script1}) {
		my $i = 1;
		while($args{'script' . $i}) {
			print "\t<script";
			
			my $scripttype = $args{'script' . $i . '_type'} if($args{'script' . $i . '_type'});
			print " type=\"$scripttype\"" if($args{'script' . $i . '_type'});
			
			my $scriptlanguage  = $args{'script' . $i . '_language'} if($args{'script' . $i . '_language'});
			print " language=\"$scriptlanguage\"" if($args{'script' . $i . '_language'});
			
			print " src=\"$args{'script' . $i}\"></script>\n";
			
			$i++;
		}
	}
	print "\t<style>$args{style}</style>\n" if($args{style});
	
	print "</head>\n\n";
	
	print "<body";
	print " onUnload=\"$args{body_onunload}\"" if($args{body_onunload});
	print " onLoad=\"$args{body_onload}\"" if($args{body_onload});
	print ">\n\n";
}

sub htmlfooter {
	print "\n</body>\n";
	print "</html>\n";
}
